
/*
 * Copyright (c) 2006-2008 by Roland Riegel <feedback@roland-riegel.de>
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "httpd_config.h"
#include "httpd_modules.h"
#include "httpd_session.h"

#include <sd/fat16.h>
#include <sd/sd.h>

#include <stdio.h>
#include <string.h>
#include <avr/pgmspace.h>

/**
 * \addtogroup app
 *
 * @{
 */
/**
 * \addtogroup app_httpd
 *
 * @{
 */
/**
 * \file
 * HTTP server module implementations (license: GPLv2)
 *
 * Current modules provide support for:
 * - HTTP Basic authorization.
 * - Custom start website.
 * - File download from the memory card.
 * - Directory listings from the memory card.
 *
 * \author Roland Riegel
 */

struct httpd_module_mime_mapping
{
    const char* file_ending;
    const char* mime;
};

static const char PROGMEM httpd_module_file_ending_css[] = ".css";
static const char PROGMEM httpd_module_file_ending_gif[] = ".gif";
static const char PROGMEM httpd_module_file_ending_html[] = ".html";
static const char PROGMEM httpd_module_file_ending_htm[] = ".htm";
static const char PROGMEM httpd_module_file_ending_jpeg[] = ".jpeg";
static const char PROGMEM httpd_module_file_ending_jpg[] = ".jpg";
static const char PROGMEM httpd_module_file_ending_js[] = ".js";
static const char PROGMEM httpd_module_file_ending_png[] = ".png";
static const char PROGMEM httpd_module_file_ending_txt[] = ".txt";
static const char PROGMEM httpd_module_file_ending_xml[] = ".xml";
static const char PROGMEM httpd_module_file_ending_zip[] = ".zip";

static const char PROGMEM httpd_module_mime_css[] = "text/css";
static const char PROGMEM httpd_module_mime_gif[] = "image/gif";
static const char PROGMEM httpd_module_mime_html[] = "text/html";
static const char PROGMEM httpd_module_mime_jpeg[] = "image/jpeg";
static const char PROGMEM httpd_module_mime_js[] = "text/javascript";
static const char PROGMEM httpd_module_mime_png[] = "image/png";
static const char PROGMEM httpd_module_mime_txt[] = "text/plain";
static const char PROGMEM httpd_module_mime_xml[] = "text/xml";
static const char PROGMEM httpd_module_mime_zip[] = "application/zip";
static const char PROGMEM httpd_module_mime_octet[] = "application/octet-stream";

struct httpd_module_mime_mapping httpd_module_mime_mappings[] PROGMEM =
{
    { httpd_module_file_ending_css,  httpd_module_mime_css   },
    { httpd_module_file_ending_gif,  httpd_module_mime_gif   },
    { httpd_module_file_ending_html, httpd_module_mime_html  },
    { httpd_module_file_ending_htm,  httpd_module_mime_html  },
    { httpd_module_file_ending_jpeg, httpd_module_mime_jpeg  },
    { httpd_module_file_ending_jpg,  httpd_module_mime_jpeg  },
    { httpd_module_file_ending_js,   httpd_module_mime_js    },
    { httpd_module_file_ending_png,  httpd_module_mime_png   },
    { httpd_module_file_ending_txt,  httpd_module_mime_txt   },
    { httpd_module_file_ending_xml,  httpd_module_mime_xml   },
    { httpd_module_file_ending_zip,  httpd_module_mime_zip   },
};

static PGM_P httpd_module_get_mime_type(const char* uri);
static void httpd_module_base64_decode(char* str);

/**
 * \internal
 * A module for handling authorization with the configured user and password.
 *
 * The authorization is checked when the module is asked if it can handle the request.
 * If the client provided an incorrect user name or password, the module takes over
 * responsibility and sends an error message back. In the successful case, the module
 * rejects responsibility and the session will be handled by another module which
 * will generate the protected content.
 */
bool httpd_module_auth_callback(struct httpd_session* session, enum httpd_module_reason reason)
{
    switch(reason)
    {
        case HTTPD_MODULE_REASON_CAN_HANDLE_REQUEST:
        {
            char* authorization = httpd_session_read_header_P(session, PSTR("Authorization"));
            if(!authorization)
                return true;
            if(strncmp_P(authorization, PSTR("Basic "), 6) != 0)
                return true;

            httpd_module_base64_decode(&authorization[6]);
            if(strcmp_P(&authorization[6], PSTR(HTTPD_AUTH_USER ":" HTTPD_AUTH_PASSWORD)) != 0)
                return true;

            return false;
        }
        case HTTPD_MODULE_REASON_HANDLE_REQUEST:
        {
            httpd_session_write_status(session, 401);
            httpd_session_write_header_PP(session, PSTR("WWW-Authenticate"), PSTR("Basic realm=\"" HTTPD_AUTH_REALM "\""));
            httpd_session_write_header_PP(session, PSTR("Content-Type"), PSTR("text/html"));
            httpd_session_begin_content_chunked(session);
            httpd_session_write_content_P(session, PSTR("401 - Unauthorized"));
            httpd_session_end_content(session);
            httpd_session_close(session);
            return true;
        }
        case HTTPD_MODULE_REASON_HANDLE_REQUEST_CONTINUE:
        {
            return false;
        }
        case HTTPD_MODULE_REASON_CLEANUP:
        {
            return true;
        }
        default:
        {
            return false;
        }
    }
}

/**
 * \internal
 * A module for providing a built-in start website.
 *
 * The module only declares itself responsible for the session if the request URI
 * equals the root path "/". In this case, a welcome web page is sent to the client.
 *
 * \note It is not checked wether the HTTP response completely fits into the
 *       transmit buffer of the connection. This would be neccessary when sending
 *       arbitrary content.
 */
bool httpd_module_root_callback(struct httpd_session* session, enum httpd_module_reason reason)
{
    switch(reason)
    {
        case HTTPD_MODULE_REASON_CAN_HANDLE_REQUEST:
        {
            return strcmp_P(session->uri, PSTR("/")) == 0;
        }
        case HTTPD_MODULE_REASON_HANDLE_REQUEST:
        case HTTPD_MODULE_REASON_HANDLE_REQUEST_CONTINUE:
        {
            char* val = httpd_session_read_argument_P(session, PSTR("val"));
            if(!val)
                return false;

            httpd_session_write_status(session, 200);
            httpd_session_write_header_PP(session, PSTR("Content-Type"), PSTR("text/html"));
            httpd_session_begin_content_chunked(session);

            httpd_session_write_content_P(session, PSTR(
                "<html><body>"
                "<h1>Welcome to mega-eth!</h1>"
                "<p>Congratulations! You have your personal version of mega-eth up and running!</p>"
                "<p>What you can do now is:</p>"
                "<ul>"
                    "<li>Browse the <a href=\"sd/\">content of the connected memory card</a>.</li>"
                    "<li>Look at a <a href=\"demo_get\">demo page which shows how to evaluate GET parameters</a>.</li>"
                    "<li>Do <a href=\"demo_post\">the same for POST parameters</a>.</li>"
                "</ul>"
                "</body></html>"
            ));

            httpd_session_end_content(session);
            httpd_session_close(session);
        }
        case HTTPD_MODULE_REASON_CLEANUP:
        {
            return true;
        }
        default:
        {
            return false;
        }
    }
}

/**
 * \internal
 * A module demonstrating usage of GET or POST arguments.
 */
bool httpd_module_demo_arg_callback(struct httpd_session* session, enum httpd_module_reason reason)
{
    switch(reason)
    {
        case HTTPD_MODULE_REASON_CAN_HANDLE_REQUEST:
        {
            if(strcmp_P(session->uri, PSTR("/demo_get")) == 0)
            {
                session->user = 0;
            }
            else if(strcmp_P(session->uri, PSTR("/demo_post")) == 0)
            {
                session->user = 1;
            }
            else
            {
                return false;
            }

            return true;
        }
        case HTTPD_MODULE_REASON_HANDLE_REQUEST:
        case HTTPD_MODULE_REASON_HANDLE_REQUEST_CONTINUE:
        {
            char* val = httpd_session_read_argument_P(session, PSTR("val"));
            if(!val)
                return false;

            PGM_P method;
            if(session->user == 0)
                method = PSTR("GET");
            else
                method = PSTR("POST");

            httpd_session_write_status(session, 200);
            httpd_session_write_header_PP(session, PSTR("Content-Type"), PSTR("text/html"));
            httpd_session_begin_content_chunked(session);

            httpd_session_write_content_P(session, PSTR("<html><body><h1>mega-eth "));
            httpd_session_write_content_P(session, method);
            httpd_session_write_content_P(session, PSTR(" demo</h1>"));

            httpd_session_write_content_P(session, PSTR("<p><i>"));
            httpd_session_write_content(session, (uint8_t*) val, strlen(val));
            httpd_session_write_content_P(session, PSTR("</i></p>"));

            httpd_session_write_content_P(session, PSTR("<p><form action=\""));
            httpd_session_write_content(session, (uint8_t*) session->uri, strlen(session->uri));
            httpd_session_write_content_P(session, PSTR("\" method=\""));
            httpd_session_write_content_P(session, method);
            httpd_session_write_content_P(session, PSTR("\"><textarea name=\"val\" cols=\"50\" rows=\"10\">"));
            httpd_session_write_content(session, (uint8_t*) val, strlen(val));
            httpd_session_write_content_P(session, PSTR("</textarea><br><input type=\"submit\" value=\""));
            httpd_session_write_content_P(session, method);
            httpd_session_write_content_P(session, PSTR("\"></form></p>"));

            httpd_session_write_content_P(session, PSTR("</body></html>"));

            httpd_session_end_content(session);
            httpd_session_close(session);
        }
        case HTTPD_MODULE_REASON_CLEANUP:
        {
            return true;
        }
        default:
        {
            return false;
        }
    }
}

/**
 * \internal
 * A module for providing directory listings.
 *
 * If the request URI corresponds to a directory on the connected memory card,
 * it generates a listing with each entry corresponding to a file or a subdirectory
 * of that directory.
 *
 * For directory listing requests without final slash (e.g. "/directory"
 * instead of "/directory/") the HTTP client gets forwarded to the URL including
 * the slash.
 */
bool httpd_module_dir_callback(struct httpd_session* session, enum httpd_module_reason reason)
{
    struct fat16_dir_struct* dd = (struct fat16_dir_struct*) session->user;
    struct fat16_dir_entry_struct dir_entry;

    switch(reason)
    {
        case HTTPD_MODULE_REASON_CAN_HANDLE_REQUEST:
        {
            if(strncmp_P(session->uri, PSTR("/sd/"), 4) != 0)
                return false;
            if(!fat16_get_dir_entry_of_path(sd_get_fs(), session->uri + 3, &dir_entry))
                return false;
            if(!(dir_entry.attributes & FAT16_ATTRIB_DIR))
                return false;

            uint8_t uri_len = strlen(session->uri);
            if(session->uri[uri_len - 1] != '/')
            {
                /* this is dirty, but works for now */
                session->uri[uri_len] = '/';
                session->uri[uri_len + 1] = '\0';

                httpd_session_write_status(session, 302);
                httpd_session_write_header_P(session, PSTR("Location"), session->uri);
                httpd_session_begin_content(session);
                httpd_session_end_content(session);
                httpd_session_close(session);
                return true;
            }

            session->user = (uint16_t) fat16_open_dir(sd_get_fs(), &dir_entry);
            dd = (struct fat16_dir_struct*) session->user;

            if(!dd)
            {
                httpd_session_write_error(session, 404);
                httpd_session_close(session);
                return true;
            }

            httpd_session_write_status(session, 200);
            httpd_session_write_header_PP(session, PSTR("Content-Type"), PSTR("text/html"));
            httpd_session_begin_content(session);
            httpd_session_write_content_P(session, PSTR("<html><h1>Directory listing of "));
            httpd_session_write_content(session, (uint8_t*) session->uri, uri_len);
            httpd_session_write_content_P(session, PSTR("</h1><body><pre><table>"));
            return true;
        }
        case HTTPD_MODULE_REASON_HANDLE_REQUEST:
        case HTTPD_MODULE_REASON_HANDLE_REQUEST_CONTINUE:
        {
            while(1)
            {
                char* buffer = (char*) httpd_session_get_write_buffer(session);
                uint16_t space = httpd_session_get_write_buffer_size(session);
                if(space < 2 * sizeof(dir_entry.long_name) + 42 + 10)
                    return true;
                if(!buffer)
                    break;

                if(!fat16_read_dir(dd, &dir_entry))
                    break;

                uint8_t written = sprintf_P(buffer, PSTR("<tr><td>"));
                PGM_P format;
                if(dir_entry.attributes & FAT16_ATTRIB_DIR)
                    format = PSTR("<a href=\"%s/\">%s/</a>");
                else
                    format = PSTR("<a href=\"%s\">%s</a>");
                written += sprintf_P(buffer + written, format, dir_entry.long_name, dir_entry.long_name);
                written += sprintf_P(buffer + written, PSTR("</td><td>%lu</td></tr>"), dir_entry.file_size);

                if(httpd_session_reserve_write_buffer(session, written) != written)
                    break;
            }

            httpd_session_write_content_P(session, PSTR("</table></pre></body></html>"));
            httpd_session_end_content(session);
            httpd_session_close(session);
            return true;
        }
        case HTTPD_MODULE_REASON_CLEANUP:
        {
            fat16_close_dir(dd);
            return true;
        }
        default:
        {
            return false;
        }
    }
}

/**
 * \internal
 * A module for providing file access.
 *
 * If the request URI corresponds to a file on the connected memory card, it responds
 * with the contents of that file. That is, it actually enables the client to download
 * files from the card.
 *
 * The size of the file is included in a Content-Length header, informing the HTTP
 * client of the size of the transfer. This also enables it to display a progress bar
 * when downloading the file.
 *
 * The file extension is used to look up the mime-type of the file, which is then sent
 * in a Content-Type header and thus tells the HTTP client how to handle the file.
 */
bool httpd_module_file_callback(struct httpd_session* session, enum httpd_module_reason reason)
{
    struct fat16_file_struct* fd = (struct fat16_file_struct*) session->user;

    switch(reason)
    {
        case HTTPD_MODULE_REASON_CAN_HANDLE_REQUEST:
        {
            if(strncmp_P(session->uri, PSTR("/sd/"), 4) != 0)
                return false;

            struct fat16_dir_entry_struct dir_entry;
            if(fat16_get_dir_entry_of_path(sd_get_fs(), session->uri + 3, &dir_entry))
            {
                session->user = (uint16_t) fat16_open_file(sd_get_fs(), &dir_entry);
            }
            else
            {
                session->user = 0;
            }

            fd = (struct fat16_file_struct*) session->user;

            if(!fd)
            {
                httpd_session_write_error(session, 404);
                httpd_session_close(session);
                return true;
            }

            char length_string[12];
            sprintf_P(length_string, PSTR("%lu"), dir_entry.file_size);

            httpd_session_write_status(session, 200);
            httpd_session_write_header_P(session, PSTR("Content-Length"), length_string);
            httpd_session_write_header_PP(session, PSTR("Content-Type"), httpd_module_get_mime_type(session->uri));
            httpd_session_begin_content(session);
            return true;
        }
        case HTTPD_MODULE_REASON_HANDLE_REQUEST:
        case HTTPD_MODULE_REASON_HANDLE_REQUEST_CONTINUE:
        {
            do
            {
                uint8_t* buffer = httpd_session_get_write_buffer(session);
                uint16_t space = httpd_session_get_write_buffer_size(session);
                if(!buffer || space < 1)
                    break;

                int16_t bytes_read = fat16_read_file(fd, buffer, space);
                if(bytes_read <= 0)
                    break;

                if(httpd_session_reserve_write_buffer(session, bytes_read) != bytes_read)
                    break;

                if(bytes_read < space)
                    break;

                return true;

            } while(0);

            httpd_session_end_content(session);
            httpd_session_close(session);
            return true;
        }
        case HTTPD_MODULE_REASON_CLEANUP:
        {
            fat16_close_file(fd);
            return true;
        }
        default:
        {
            return false;
        }
    }
}

/**
 * Determines the mime-type of a HTTP transfer by the file extension.
 *
 * Searches a list for a matching file extension and returns the
 * corresponding mime-type.
 *
 * \param[in] uri The filename or URI for which to determine the mime-type.
 * \returns A pointer to the mime-type string in program memory space.
 */
PGM_P httpd_module_get_mime_type(const char* uri)
{
    for(uint8_t i = 0; i < sizeof(httpd_module_mime_mappings) / sizeof(httpd_module_mime_mappings[0]); ++i)
    {
        PGM_P ending = (PGM_P) pgm_read_word_near(&httpd_module_mime_mappings[i].file_ending);
        uint8_t ending_length = strlen_P(ending);
        uint16_t uri_length = strlen(uri);
        if(uri_length < ending_length)
            continue;
        if(strcmp_P(uri + uri_length - ending_length, ending) != 0)
            continue;

        return (PGM_P) pgm_read_word_near(&httpd_module_mime_mappings[i].mime);
    }

    return httpd_module_mime_octet;
}

/**
 * Decodes a Base64-encoded string.
 *
 * \param[in,out] str The encoded string which gets decoded.
 */
void httpd_module_base64_decode(char* str)
{
    char* out = str;
    char stream[4];
    while(strlen(str) >= 4)
    {
        for(uint8_t i = 0; i < 4; ++i)
        {
            if(*str >= 'A' && *str <= 'Z')
            {
                stream[i] = *str - 'A';
            }
            else if(*str >= 'a' && *str <= 'z')
            {
                stream[i] = *str - 'a' + 26;
            }
            else if(*str >= '0' && *str <= '9')
            {
                stream[i] = *str - '0' + 52;
            }
            else if(*str == '+')
            {
                stream[i] = 62;
            }
            else if(*str == '/')
            {
                stream[i] = 63;
            }
            else if(*str == '=')
            {
                stream[i] = 0;
            }

            ++str;
        }

        *out++ = stream[0] << 2 | stream[1] >> 4;
        *out++ = stream[1] << 4 | stream[2] >> 2;
        *out++ = stream[2] << 6 | stream[3] >> 0;
    }

    *out = '\0';
}

/**
 * @}
 * @}
 */

