
/*
 * Copyright (c) 2006-2008 by Roland Riegel <feedback@roland-riegel.de>
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <avr/io.h>
#include <util/delay.h>

#include "enc28j60_config.h"
#include "enc28j60_io.h"

#include <arch/spi.h>

/**
 * \addtogroup net
 *
 * @{
 */
/**
 * \addtogroup net_driver
 *
 * @{
 */
/**
 * \addtogroup net_driver_enc28j60
 *
 * @{
 */
/**
 * \file
 * Microchip ENC28J60 I/O implementation (license: GPLv2)
 *
 * \author Roland Riegel
 */

#define enc28j60_rcr(address) spi_send_byte(0x00 | (address))
#define enc28j60_rbm() spi_send_byte(0x3a)
#define enc28j60_wcr(address) spi_send_byte(0x40 | (address))
#define enc28j60_wbm(address) spi_send_byte(0x7a)
#define enc28j60_bfs(address) spi_send_byte(0x80 | (address))
#define enc28j60_bfc(address) spi_send_byte(0xa0 | (address))
#define enc28j60_sc() spi_send_byte(0xff)

static void enc28j60_bank(uint8_t num);

/**
 * \internal
 * Initializes the SPI interface to the ENC28J60 chips.
 */
void enc28j60_io_init()
{
    /* configure pins */
    enc28j60_config_io();
}

/**
 * \internal
 * Forces a reset to the ENC28J60.
 *
 * After the reset a reinitialization is necessary.
 */
void enc28j60_reset()
{
    enc28j60_reset_on();
    _delay_ms(10);
    enc28j60_reset_off();
    _delay_ms(10);
}

/**
 * \internal
 * Reads the value of a hardware register.
 *
 * \param[in] address The address of the register to read.
 * \returns The register value.
 */
uint8_t enc28j60_read(uint8_t address)
{
    uint8_t result;

    /* switch to register bank */
    enc28j60_bank((address & MASK_BANK) >> SHIFT_BANK);
    
    /* read from address */
    enc28j60_select();
    enc28j60_rcr((address & MASK_ADDR) >> SHIFT_ADDR);
    if(address & MASK_DBRD)
        spi_rec_byte();
    result = spi_rec_byte();
    enc28j60_deselect();

    return result;
}

/**
 * \internal
 * Writes the value of a hardware register.
 *
 * \param[in] address The address of the register to write.
 * \param[in] value The value to write into the register.
 */
void enc28j60_write(uint8_t address, uint8_t value)
{
    /* switch to register bank */
    enc28j60_bank((address & MASK_BANK) >> SHIFT_BANK);
    
    /* write to address */
    enc28j60_select();
    enc28j60_wcr((address & MASK_ADDR) >> SHIFT_ADDR);
    spi_send_byte(value);
    enc28j60_deselect();
}

/**
 * \internal
 * Clears bits in a hardware register.
 *
 * Performs a NAND operation on the current register value
 * and the given bitmask.
 *
 * \param[in] address The address of the register to alter.
 * \param[in] bits A bitmask specifiying the bits to clear.
 */
void enc28j60_clear_bits(uint8_t address, uint8_t bits)
{
    /* switch to register bank */
    enc28j60_bank((address & MASK_BANK) >> SHIFT_BANK);
    
    /* write to address */
    enc28j60_select();
    enc28j60_bfc((address & MASK_ADDR) >> SHIFT_ADDR);
    spi_send_byte(bits);
    enc28j60_deselect();
}

/**
 * \internal
 * Sets bits in a hardware register.
 *
 * Performs an OR operation on the current register value
 * and the given bitmask.
 *
 * \param[in] address The address of the register to alter.
 * \param[in] bits A bitmask specifiying the bits to set.
 */
void enc28j60_set_bits(uint8_t address, uint8_t bits)
{
    /* switch to register bank */
    enc28j60_bank((address & MASK_BANK) >> SHIFT_BANK);
    
    /* write to address */
    enc28j60_select();
    enc28j60_bfs((address & MASK_ADDR) >> SHIFT_ADDR);
    spi_send_byte(bits);
    enc28j60_deselect();
}

/**
 * \internal
 * Reads the value of a hardware PHY register.
 *
 * \param[in] address The address of the PHY register to read.
 * \returns The register value.
 */
uint16_t enc28j60_read_phy(uint8_t address)
{
    enc28j60_write(MIREGADR, address);
    enc28j60_set_bits(MICMD, (1 << MICMD_MIIRD));

    _delay_us(10);

    while(enc28j60_read(MISTAT) & (1 << MISTAT_BUSY));
    
    enc28j60_clear_bits(MICMD, (1 << MICMD_MIIRD));

    return ((uint16_t) enc28j60_read(MIRDH)) << 8 |
           ((uint16_t) enc28j60_read(MIRDL));
}

/**
 * \internal
 * Writes the value to a hardware PHY register.
 *
 * \param[in] address The address of the PHY register to write.
 * \param[in] value The value to write into the register.
 */
void enc28j60_write_phy(uint8_t address, uint16_t value)
{
    enc28j60_write(MIREGADR, address);
    enc28j60_write(MIWRL, value & 0xff);
    enc28j60_write(MIWRH, value >> 8);
    
    _delay_us(10);

    while(enc28j60_read(MISTAT) & (1 << MISTAT_BUSY));
}

/**
 * \internal
 * Reads a byte from the RAM buffer at the current position.
 *
 * \returns The byte read from the current RAM position.
 */
uint8_t enc28j60_read_buffer_byte()
{
    uint8_t b;

    enc28j60_select();
    enc28j60_rbm();

    b = spi_rec_byte();
    
    enc28j60_deselect();

    return b;
}

/**
 * \internal
 * Writes a byte to the RAM buffer at the current position.
 *
 * \param[in] b The data byte to write.
 */
void enc28j60_write_buffer_byte(uint8_t b)
{
    enc28j60_select();
    enc28j60_wbm();

    spi_send_byte(b);
    
    enc28j60_deselect();
}

/**
 * \internal
 * Reads multiple bytes from the RAM buffer.
 *
 * \param[out] buffer A pointer to the buffer which receives the data.
 * \param[in] buffer_len The buffer length and number of bytes to read.
 */
void enc28j60_read_buffer(uint8_t* buffer, uint16_t buffer_len)
{
    enc28j60_select();
    enc28j60_rbm();

    spi_rec_data(buffer, buffer_len);
    
    enc28j60_deselect();
}

/**
 * \internal
 * Writes multiple bytes to the RAM buffer.
 *
 * \param[in] buffer A pointer to the buffer containing the data to write.
 * \param[in] buffer_len The number of bytes to write.
 */
void enc28j60_write_buffer(uint8_t* buffer, uint16_t buffer_len)
{
    enc28j60_select();
    enc28j60_wbm();

    spi_send_data(buffer, buffer_len);
    
    enc28j60_deselect();
}

/**
 * Switches the hardware register bank.
 *
 * \param[in] num The index of the register bank to switch to.
 */
void enc28j60_bank(uint8_t num)
{
    static uint8_t bank = 0xff;

    if(num == bank)
        return;
    
    /* clear bank bits */
    enc28j60_select();
    enc28j60_bfc(ECON1);
    spi_send_byte(0x03);
    enc28j60_deselect();

    /* set bank bits */
    enc28j60_select();
    enc28j60_bfs(ECON1);
    spi_send_byte(num);
    enc28j60_deselect();

    bank = num;
}

/**
 * @}
 * @}
 * @}
 */

